/*
 * CComplex.cpp
 *
 *  Created on: Dec 6, 2013
 *      Author: Jono
 */

#include "CComplex.h"


/*
***************************
******   PRIVATE    *******
*************************** 
*/

/*
The function "element_of" checks to see if the list contains the given element.
*/
bool CComplex::element_of(int element, std::vector <unsigned int > list ){
  if( list.size() == 0){
    std::cout << "ERROR: Empty List \n";
    return false;
  }
  else{
    for(int i = 0 ; i< list.size() ; i++){
      if ( list[i] == element )
        return true;
    }
    return false;
  }
};

/* 
This function infers which (nonzero-dimensional) faces to include based on which verticies need to be included.

The input is a boolean vector of length 4.
The output is a boolean vector of length 9.

Recall that 
the 2-cell comes first
the edges are ordered counter clockwise, starting at the bottom
the positions of the input verticies are given by :: (0,0) ; (0,1) ; (1,0) ; (1,1)  

This is governed by the following fact::
An analytically verified cell is in the approximation if and only if all of its verticies are in the approximation

In this manner, we have the following relations:
v_0 && v_2 ==> e_0
v_2 && v_3 ==> e_1
v_1 && v_3 ==> e_2
v_0 && v_1 ==> e_3
*/  
std::vector< bool >  CComplex::infer_faces(  std::vector< bool >  vertex ){
  std::vector< bool > output;
  /* We handle the 2-cell first */
  if( vertex[0] && vertex[1] && vertex[2] && vertex[3] )
    output.push_back(1);
  else
    output.push_back(0);
  
  /* Next, we infer which edges to include */ 
  
//   v_0 && v_2 ==> e_0
  if( vertex[0] && vertex[2] )
    output.push_back(1);
  else
    output.push_back(0);
  
//   v_2 && v_3 ==> e_1
  if( vertex[2] && vertex[3] )
    output.push_back(1);
  else
    output.push_back(0);
  
//   v_1 && v_3 ==> e_2
  if( vertex[1] && vertex[3] )
    output.push_back(1);
  else
    output.push_back(0);
  
//   v_0 && v_1 ==> e_3
  if( vertex[0] && vertex[1] )
    output.push_back(1);
  else
    output.push_back(0);
  
  /* We now add the vertices to the list */
  for(int i =0; i <4 ; i++){
    output.push_back(vertex[i]);
  }
  
  return output;
}




/*
Returns a boolean < True > if we've aquired the total cover. A boolean false if we did not. 

We split this function up into two cases:: The face is an edge or the face is a vertex.
*/
bool CComplex::total_cover_test(std::vector <unsigned int > splitting_path , int face){
  if( face < 0 || face > 8){
    std::cout << " Unexpected input in total_cover_test \n";
    return false;
  }
  /* If we receive the 2-cell, then that is a total cover */
  else if( face == 0 ){
    std::cout << "The function << total_cover_test >> received a 2-cell \n";
    return true;
  }  
  /* First we suppose that the face is an edge */
  else if( face < 5){
    /* First suppose that we are dealing with the << BOTTOM >> edge */
    if ( face == 1){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(2,splitting_path) || element_of(3,splitting_path) )
        return true;
      else
        return false;
    }
    /* Next suppose that we are dealing with the << RIGHT >>  edge */
    else if ( face == 2){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(0,splitting_path) || element_of(2,splitting_path) )
        return true;
      else
        return false;
    }
    /* Next suppose that we are dealing with the << TOP >> edge */
    else if ( face == 3){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(0,splitting_path) || element_of(1,splitting_path) )
        return true;
      else
        return false;
    }
    /* Lastly, suppose that we are dealing with the << LEFT >> edge */
    else if ( face == 4){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(1,splitting_path) || element_of(3,splitting_path) )
        return true;
      else
        return false;
    }  
  }
  /* 
  Otherwise the face is a vertex 
  */
  else{
    /* First suppose that we are dealing with the << BOTTOM LEFT >> vertex        (  v_0  )*/
    if ( face == 5){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(3,splitting_path) || (element_of(1,splitting_path)  && element_of(2,splitting_path) ) )
        return true;
      else
        return false;
    }
    /* Next, suppose that we are dealing with the << TOP LEFT >> vertex           (  v_1  ) */
    if ( face == 6){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(1,splitting_path) || (element_of(0,splitting_path)  && element_of(3,splitting_path) ) )
        return true;
      else
        return false;
    }
    /* Next, suppose that we are dealing with the << BOTTOM  RIGHT >> vertex      (  v_2  ) */
    if ( face == 7){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(2,splitting_path) || (element_of(0,splitting_path)  && element_of(3,splitting_path) ) )
        return true;
      else
        return false;
    }
    /* Lastly suppose that we are dealing with the << TOP  RIGHT >> vertex        (  v_3  ) */
    if ( face == 8){
      /* if the appropriate paths were taken, then we have a total cover */
      if ( element_of(0,splitting_path) || (element_of(1,splitting_path)  && element_of(2,splitting_path) ) )
        return true;
      else
        return false;
    }
  }
}



/*
Given two splitting paths, this function determines if one of the <direction>-coordinates of the paths is an extension of the other.

direction == 0 corresponds to the x-dimension
direction == 1 corresponds to the y-dimension
*/
bool CComplex::Extension( std::vector< unsigned int >  vec_1 , std::vector< unsigned int > vec_2 , int direction){
  /* I give up on using bitwise operations */
  int min_length = std::min( vec_1.size() , vec_2.size() );
  for( int  i =0; i < min_length; i++ ){
    /* We check the X direction */ 
    if( direction == 0 ){
      if( vec_1[i] == 0 || vec_1[i] ==2){
        /* vec_1 is on the LEFT side */
        if( vec_2[i] ==1 || vec_2[i] ==3 ){
          /* The vec_2 is on the RIGHT side */
          return false;
        }
      }
      else{
        /* The vec_1 is on the RIGHT side */ 
        if(vec_2[i] ==0 || vec_2[i] ==2 ){
          /* The vec_2 is on the LEFT side */
          return false;
        }
      }
    }
    /* We check the Y direction */ 
    else if( direction == 1 ){
      if( vec_1[i] == 0 || vec_1[i] ==1){
        /* vec_1 is on the BOTTOM side */
        if( vec_2[i] ==2 || vec_2[i] ==3 ){
          /* The vec_2 is on the TOP side */
          return false;
        }
      }
      else{
        /* The vec_1 is on the TOP side */ 
        if(vec_2[i] ==0 || vec_2[i] ==1 ){
          /* The vec_2 is on the BOTTOM side */
          return false;
        }
      }
    }
    else{
      std::cout <<  " ERROR in CComplex.cpp  \n";
    }
  }
  return true;
}


/* 
Given x- and y- cells, we take their tensor product ...
*/
std::vector< int >  CComplex::Tensor_Cells( std::vector< bool>  x_cells, std::vector< bool>  y_cells){
  std::vector< int > output;
  
  if( x_cells[0] ){
    /* The LEFT vertex */
    if( y_cells[0]){
      /* The BOTTOM vertex */
      output.push_back(5);
    }
    if( y_cells[1]){
      /* The y-dimensional 1-cell */
      output.push_back(4);
    }
    if( y_cells[2]){
      /* The TOP vertex */
      output.push_back(6);
    }
  }
  if( x_cells[1] ){
    /* The x-dimensional 1-cell */
    if( y_cells[0]){
      /* The BOTTOM vertex */
      output.push_back(1);
    }
    if( y_cells[1]){
      /* The y-dimensional 1-cell */
      output.push_back(0);
    }
    if( y_cells[2]){
      /* The TOP vertex */
      output.push_back(3);
    }
  }
  if( x_cells[2] ){
    /* The RIGHT vertex */
    if( y_cells[0]){
      /* The BOTTOM vertex */
      output.push_back(7);
    }
    if( y_cells[1]){
      /* The y-dimensional 1-cell */
      output.push_back(2);
    }
    if( y_cells[2]){
      /* The TOP vertex */
      output.push_back(8);
    }
  }
  
  return output;
}



/*
This function returns a boolean vector. 
The first two values coorespond to whether the given face is degenerate in the x and y directions.  
The next two values coorespond to , whether the cell is on the left or the right side of the cube,  ( top or bottom),  given that the cell is degenerate in that direction. 
*/
std::vector < bool  > CComplex::Face_Degeneracies( int face_number_){
  std::vector < bool  >  output;

  /* We check if the x direction is degerenate */ 
  if(face_number_ == 2 || face_number_ == 4 || face_number_ >4){
    output.push_back(true);
  }
  else{
    output.push_back(false);
  }
  
  /* We now check if the y direction is degerenate */ 
  if(face_number_ == 1 || face_number_ == 3 || face_number_ >4){
    output.push_back(true);
  }
  else{
    output.push_back(false);
  }
  
  /* If the x direction of the cell is degenerate, then we determine  which side of the cube the cell is on. 
  If the cell is not degenerate in the x direction, then the third position of this vector is meaningless. */
  /*Here  left/bottom = 0 and right/top = 1*/
  if( face_number_ == 4 || face_number_ == 5 || face_number_ == 6)
    output.push_back(false);
  /* Otherwise the cell is on the right side  or at least we say it is on the right side. */ 
  else
    output.push_back(true);

  /* If the y direction of the cell is degenerate, then we determine  which side of the cube the cell is on. 
  If the cell is not degenerate in the y direction, then the fourth position of this vector is meaningless. */
  /*Here  left/bottom = 0 and right/top = 1*/
  if( face_number_ == 1 || face_number_ == 5 || face_number_ == 7)
    output.push_back(false);
  /* Otherwise the cell is on the top, or at least we say it is on top. */ 
  else
    output.push_back(true);

  return output;
}


/*  
This function iterates through each face of the given node.
After calling Redundant_Cells, we remove excessive large cells
*/
void CComplex::Prune_Node(  CNode* pt_node){
  /* It is unnecessary to look at the 2-cell */ 
  
  /* We create an empty vector, for use in the SFace structure */
  std::vector  < unsigned int > blank;
  
  for(int i = 1 ; i<9 ;i++){
    if( (*pt_node).Check_Cell( i ) ){
      /* If the vertex is attached to node, then we create an SFace and  find its redundant cells */
      SFace current_face = {blank ,pt_node,i};
      std::vector < SFace > face_list = Redundant_Cells(current_face);
      if( (i <5) && (face_list.size()>2)){
        /* If i<5, then we are dealing with a 1-cell. */
        /* If the input cell intersects at least two other cells, then the input cell is large, and should be removed */
        (*pt_node).delete_face(i); 
      }
      else{
        /* We want to remove all redundant cells expect for the input cell node */
        for( int j =0; j < face_list.size() ; j ++){
          /* We remove all  of the redundant cells, except for the original vertex */
          if( face_list[j].pt_node_  !=  pt_node){ 
            (*face_list[j].pt_node_).delete_face(face_list[j].face_number_);
          }
        }
      }
    }
  }/* end FOR */
}


/*
This function ascends to the "total cover" of a particular face.

The output is an SFace.
The splitting_path is the local splitting path of the input face.
The pointer to CNode is pointing to the total cover.
The face_number of the ouput cooresponds to the entire cube


Again, The path member of the output may not be its complete path, 
but rather describes how to descend the quad tree from the output node down to the input node.

*/
SFace CComplex::Ascend(CNode * in_node, int face){
	/* We initialize the output */
	SFace output;
	/* The face_number of the ouput cooresponds to the entire cube */
	output.face_number_ = 0;
	/* We initialize a vector which allows us to keep track of how we are ascending the tree */
	std::vector < unsigned int > splitting_path;
	/* If the input node is the root, we return the root immediately */
	if (in_node == &root){
		output.pt_node_ = in_node;
	}
	else{
		/* We initalize a pointer->node called < pt_child >   */
		CNode * pt_child = in_node;
		/* We continue to ascend the tree until we reach the root. */
		while (!((pt_child == &root))){
			/* We compute the birth order of the child, and add it to the splitting path */
			/*WARNING This way of constructing the splitting path for a node is reversed from the normal method of construction WARNING  */
			splitting_path.push_back((*pt_child).child_number());
			/* We set the child equal to its parent */
			pt_child = (*pt_child).get_parent();
			/* If we have obtained a total cover before reaching the root, we break out of the while loop */
			if (total_cover_test(splitting_path, face))
				break;
		}
		output.pt_node_ = pt_child;
	}
	/* We set output.path_ to be equal to splitting_path , in reverse order */
	while (!(splitting_path.empty())){
		output.path_.push_back(splitting_path.back());
		splitting_path.pop_back();
	}

	return output;
}




/*
Obtains a list of SFaces which have a nonempty intersection with the input SFace.
Outputs the correct splitting path.

the SFaces need not be in the complex.

Assumes in_face has correct splitting path relative to total_cover

Possible Errors:: A node my have no children and no face info. This will result in a seg fault.
*/
std::vector < SFace > CComplex::Redundant_Cells_Assisted(SFace in_face, SFace total_cover){

	/* We initialize our output */
	std::vector < SFace > output;
	/* We initialize our list of cubes to check */
	std::vector < SFace > cubes_to_check;
	/* We add the total cover to the list of nodes to check */
	cubes_to_check.push_back(total_cover);

	/* We descend through the tree until we obtain all of the cells which intersect our input cell */
	while (!(cubes_to_check.empty())){
		/* We set Current_Node to be the last node added to the list of nodes we must check */
		SFace Current_Cube = cubes_to_check.back();
		/* We remove Current_Node from the list of nodes to check */
		cubes_to_check.pop_back();

		/*  If Current_Node is a leaf, then we check to see if its faces intersect the input face */
		if ((*Current_Cube.pt_node_).leaf_check()){
			/* We gather a list of all faces from Current_Cube which intersect our input face */
			std::vector< int >  face_list = Intersecting_Faces(Current_Cube, in_face);
			/* We add all of the cells from face_list to the output */
			for (int i = 0; i < face_list.size(); i++){
				Current_Cube.face_number_ = face_list[i];
				output.push_back(Current_Cube);
			}
		}
		else{
			/* If the cube is not a leaf we add the children of Current_Node to the list of nodes to check */
			for (int i = 0; i < 4; i++){
				/* We obtain a child of Current_Cube. (note, its face number doesn't matter) */
				SFace child = { Current_Cube.path_, (*(Current_Cube.pt_node_)).get_child(i), 0 };

				/* We update the child's splitting path */
				child.path_.push_back(i);

				/*  If the child cube of Current_Cube intersects the input face, then we add it to the output. */
				if (Intersect_Cube(child, in_face)){
					cubes_to_check.push_back(child);
				}
			}
		}
	}
	return output;

}



/*
***************************
******  PROTECTED   *******
***************************
*/



/*
Given a cube WHICH INTERSECTS THE INPUT CELL, this function returns the faces of that cube which have a nontrivial intersection with the input cell.
*/
std::vector < int >  CComplex::Intersecting_Faces(SFace cube, SFace cell){


	/* We determine if the cell is degerate in the x and y directions, and store these values  */
	bool degenerate_x, degenerate_y;
	std::vector < bool > degeneracies = Face_Degeneracies(cell.face_number_);
	degenerate_x = degeneracies[0];
	degenerate_y = degeneracies[1];
	/* if the cell is degenerate in some direction, we need to determine which side of the cube the cell is on.
	Here  left/bottom = 0 and right/top = 1*/
	bool side_x, side_y;
	side_x = degeneracies[2];
	side_y = degeneracies[3];

	/* We determine if the x- or y- address of cube or cell is an extension of the other */
	bool x_extension = Extension(cube.path_, cell.path_, 0);
	bool y_extension = Extension(cube.path_, cell.path_, 1);

	/* We need to create some type of tensor product.  We initialize the facotrs of this tensor product. */
	std::vector< bool> x_face_factors;
	std::vector< bool> y_face_factors;
	for (int i = 0; i <3; i++){
		x_face_factors.push_back(0);
		y_face_factors.push_back(0);
	}

	/* We determine which x-face factors we need */
	if (x_extension){
		/* Either the cell or the cube is an extension of the other */
		if (degenerate_x){
			/* The cell is degenerate in the x direction */
			if (cube.path_.size() < cell.path_.size()){
				/* the cell is smaller than the cube  */
				/* We check if the cell is in the interior of the cube */
				bool extension_has_1 = false;
				bool extension_has_0 = false;
				/* We check if there is a '1' in cube's address, after cell's address part */
				for (int i = 0; i < (cell.path_.size() - cube.path_.size()); i++){
					if (cell.path_[i + cube.path_.size()] == 1 || cell.path_[i + cube.path_.size()] == 3){
						extension_has_1 = true;
						break;
					}
				}
				/* We check if there is a '0' in cube's address, after cell's address part */
				for (int i = 0; i < (cell.path_.size() - cube.path_.size()); i++){
					if (cell.path_[i + cube.path_.size()] == 0 || cell.path_[i + cube.path_.size()] == 2){
						extension_has_0 = true;
						break;
					}
				}
				if (extension_has_0 && extension_has_1){
					/* The cell is in the interior of the cube */
					x_face_factors[1] = 1;
				}
				else{
					/* The cell is not in the interior of the cube */
					if (!(extension_has_0)){
						/* The cell is flush with the RIGHT  side of the cube */
						if (side_x){
							/* The vertex is on the RIGHT  side of the cell, we return the right vertex  */
							x_face_factors[2] = 1;
						}
						else{
							/* We return the 1-cell */
							x_face_factors[1] = 1;
						}
					}
					else{
						/* The cell is flush with the LEFT  side of the cube */
						if (side_x){
							/* The vertex is on the RIGHT  side of the cell, we return the 1-cell  */
							x_face_factors[1] = 1;
						}
						else{
							/* We return the LEFT  vertex */
							x_face_factors[0] = 1;
						}
					}
				}
			}
			else{
				/* the cell is the same size or bigger than the cube */
				if (side_x){
					/* the face is on the RIGHT side of the cell, return RIGHT vertex */
					x_face_factors[2] = 1;
				}
				else{
					/* the face is on the LEFT side of the cell, return LEFT vertex */
					x_face_factors[0] = 1;
				}
			}
		}
		else{
			/* The cell is non-degenerate in the x direction, add the 1 cell to the output */
			x_face_factors[1] = 1;
			if (cube.path_.size() > cell.path_.size()){
				bool extension_has_1 = false;
				bool extension_has_0 = false;
				/* We check if there is a '1' in cube's address, after cell's address part */
				for (int i = 0; i < (cube.path_.size() - cell.path_.size()); i++){
					if (cube.path_[i + cell.path_.size()] == 1 || cube.path_[i + cell.path_.size()] == 3){
						extension_has_1 = true;
						break;
					}
				}
				/* We check if there is a '0' in cube's address, after cell's address part */
				for (int i = 0; i < (cube.path_.size() - cell.path_.size()); i++){
					if (cube.path_[i + cell.path_.size()] == 0 || cube.path_[i + cell.path_.size()] == 2){
						extension_has_0 = true;
						break;
					}
				}
				if (extension_has_1){
					/* We add the left vertex to the output pile */
					x_face_factors[0] = 1;
				}
				if (extension_has_0){
					/* We add the RIGHT vertex to the output pile */
					x_face_factors[2] = 1;
				}
			}
		}
	}
	else{
		/* The cell and the cube are in separate cubes. */
		if (degenerate_x){
			if (side_x){
				/* The cell is on the right side, return the left vertex */
				x_face_factors[0] = 1;
			}
			else{
				/* The cell is on the LEFT side, return the RIGHT vertex */
				x_face_factors[2] = 1;
			}
		}
	}

	/* Again, I am shamelessly copying the previous 100+ lines of code and changing the x-s to y-s. */

	/* We determine which y-face factors we need */
	if (y_extension){
		/* Either the cell or the cube is an extension of the other */
		if (degenerate_y){
			/* The cell is degenerate in the y direction */
			if (cube.path_.size() < cell.path_.size()){
				/* the cell is smaller than the cube  */
				/* We check if the cell is in the interior of the cube */
				bool extension_has_1 = false;
				bool extension_has_0 = false;
				/* We check if there is a '1' in cube's address, after cell's address part */
				for (int i = 0; i < (cell.path_.size() - cube.path_.size()); i++){
					if (cell.path_[i + cube.path_.size()] == 2 || cell.path_[i + cube.path_.size()] == 3){
						extension_has_1 = true;
						break;
					}
				}
				/* We check if there is a '0' in cube's address, after cell's address part */
				for (int i = 0; i < (cell.path_.size() - cube.path_.size()); i++){
					if (cell.path_[i + cube.path_.size()] == 0 || cell.path_[i + cube.path_.size()] == 1){
						extension_has_0 = true;
						break;
					}
				}
				if (extension_has_0 && extension_has_1){
					/* The cell is in the interior of the cube */
					y_face_factors[1] = 1;
				}
				else{
					/* The cell is not in the interior of the cube */
					if (!(extension_has_0)){
						/* The cell is flush with the TOP  side of the cube */
						if (side_y){
							/* The vertex is on the TOP  side of the cell, we return the TOP vertex  */
							y_face_factors[2] = 1;
						}
						else{
							/* We return the 1-cell */
							y_face_factors[1] = 1;
						}
					}
					else{
						/* The cell is flush with the BOTTOM  side of the cube */
						if (side_y){
							/* The vertex is on the TOP  side of the cell, we return the 1-cell  */
							y_face_factors[1] = 1;
						}
						else{
							/* We return the BOTTOM  vertex */
							y_face_factors[0] = 1;
						}
					}
				}
			}
			else{
				/* the cell is the same size or bigger than the cube */
				if (side_y){
					/* the face is on the TOP side of the cell, return TOP vertex */
					y_face_factors[2] = 1;
				}
				else{
					y_face_factors[0] = 1;
				}
			}
		}
		else{
			/* The cell is non-degenerate in the y direction, add the 1 cell to the output */
			y_face_factors[1] = 1;
			if (cube.path_.size() > cell.path_.size()){
				bool extension_has_1 = false;
				bool extension_has_0 = false;
				/* We check if there is a '1' in cube's address, after cell's address part */
				for (int i = 0; i < (cube.path_.size() - cell.path_.size()); i++){
					if (cube.path_[i + cell.path_.size()] == 2 || cube.path_[i + cell.path_.size()] == 3){
						extension_has_1 = true;
						break;
					}
				}
				/* We check if there is a '0' in cube's address, after cell's address part */
				for (int i = 0; i < (cube.path_.size() - cell.path_.size()); i++){
					if (cube.path_[i + cell.path_.size()] == 0 || cube.path_[i + cell.path_.size()] == 1){
						extension_has_0 = true;
						break;
					}
				}
				if (extension_has_1){
					/* We add the BOTTOM vertex to the output pile */
					y_face_factors[0] = 1;
				}
				if (extension_has_0){
					/* We add the TOP vertex to the output pile */
					y_face_factors[2] = 1;
				}
			}
		}
	}
	else{
		/* The cell and the cube are in separate cubes. */
		if (degenerate_y){
			if (side_y){
				/* The cell is on the TOP side, return the BOTTOM vertex */
				y_face_factors[0] = 1;
			}
			else{
				/* The cell is on the BOTTOM side, return the TOP vertex */
				y_face_factors[2] = 1;
			}
		}
	}

	/* We initialize the output */
	std::vector < int > output = Tensor_Cells(x_face_factors, y_face_factors);
	return output;
}

/*
This function returns a boolean value indicating whether a closed cube intersects a given cell.
*/
bool CComplex::Intersect_Cube(SFace cube, SFace cell){
	/* If the cube is the root, then it definately contains the cell. */
	if (cube.path_.size() == 0)
		return true;

	/* We determine if the cell is degerate in the x and y directions, and store these values  */
	bool degenerate_x, degenerate_y;
	std::vector < bool > degeneracies = Face_Degeneracies(cell.face_number_);
	degenerate_x = degeneracies[0];
	degenerate_y = degeneracies[1];
	/* if the cell is degenerate in some direction, we need to determine which side of the cube the cell is on.
	Here  left/bottom = 0 and right/top = 1*/
	bool side_x, side_y;
	side_x = degeneracies[2];
	side_y = degeneracies[3];

	/*  We create the boolean variables outside_x and outside_y which are indicative of whether the cube's splitting path index no longer aggrees with the splitting path index of the cell. */
	bool outside_x = false;
	bool outside_y = false;
	/* We need to iterate to the end of the longest splitting path (between the cube and the cell). */
	int long_path = std::max(cube.path_.size(), cell.path_.size());
	for (int i = 0; i< long_path; i++){
		/* We iterate through each dimension of the cube */
		/*  We do this manually for the time being .  We will need to implement tthis using a for loop later .*/
		if (!(outside_x)){
			/* We differentiate between the two cases, where the path index is smaller than the cell, and where the path index is the same size or bigger than the cell */
			if (i + 1 > cell.path_.size()){
				/*  If the path index is smaller than the cell, then we test whether the cell is degenerate in the x-direction. */
				if (degenerate_x){
					/* The cell is x degenerate . We now test which side of the cube the cell is on  */
					if (side_x){
						/* the cell is on the right side */
						/* If the cube path is to the left,  then we return false */
						if (cube.path_[i] == 0 || cube.path_[i] == 2)
							return false;
					}
					else{
						/* the cell is on the left side */
						/* If the cube path is to the right,  then we return false */
						if (cube.path_[i] == 1 || cube.path_[i] == 3)
							return false;
					}
				}
				/* If the cell is not degenerate in the x direction, then we continue, and now test the y-direction */
			}
			/* Otherwise the path index  is the same size or bigger than the cell */
			else if (i < cube.path_.size()){
				/* path index is comparable between the cube and the cel */
				/* We now test to see if the cube path and the cell path are on opposite x-sides */
				bool different_sides;
				if (cube.path_[i] == 0 || cube.path_[i] == 2){
					/* The cube is on the left side */
					if (cell.path_[i] == 1 || cell.path_[i] == 3){
						/* The cell is on the right side */
						different_sides = true;
					}
					else
						different_sides = false;
				}
				else{
					/* The cube is on the right side */
					if (cell.path_[i] == 1 || cell.path_[i] == 3){
						/* The cell is on the right side */
						different_sides = false;
					}
					else
						different_sides = true;
				}
				/* If the cube and the cell are on the same side, then we just continue. */
				if (different_sides){
					/* We check to see if the cell is degenerate in the x direction */
					if (degenerate_x){
						outside_x = true;
						/* We check which side of the cube the cell is on */
						if (side_x){
							/* We are on the RIGHT side */
							/* If the cell.path is on the RIGHT side, then we return false */
							if (cell.path_[i] == 1 || cell.path_[i] == 3)
								return false;
						}
						else{
							/* We are on the LEFT side */
							/* If the cell.path is on the LEFT side, then we return false */
							if (cell.path_[i] == 0 || cell.path_[i] == 2)
								return false;
						}
					}
					else{
						return false;
					}
				}
			}
		}
		/* We now handle the case where the cube is outside the path of the cell. */
		else{
			/* We test which side of the cube the cell is on  */
			if (side_x){
				/* the cell is on the RIGHT side */
				if (i < cube.path_.size()){
					/* If the cube path is to the RIGHT,  then we return false , otherwise continue*/
					if (cube.path_[i] == 1 || cube.path_[i] == 3)
						return false;
				}
				if (i < cell.path_.size()){
					/* If the cell path is to the LEFT,  then we return false , otherwise continue*/
					if (cell.path_[i] == 0 || cell.path_[i] == 2)
						return false;
				}
			}
			else{
				/* the cell is on the LEFT side  */
				if (i < cube.path_.size()){
					/* If the cube path is to the LEFT,  then we return false , otherwise continue*/
					if (cube.path_[i] == 0 || cube.path_[i] == 2)
						return false;
				}
				if (i < cell.path_.size()){
					/* If the cell path is to the RIGHT,  then we return false , otherwise continue*/
					if (cell.path_[i] == 1 || cell.path_[i] == 3)
						return false;
				}
			}
		}
		/*
		In a complete breach of coding etiquette, I am essentially copying the previous 100 lines of code, and changing all of the x-s to y-s

		I am very sorry.
		*/

		if (!(outside_y)){
			/* We differentiate between the two cases, where the path index is smaller than the cell, and where the path index is the same size or bigger than the cell */
			if (i + 1> cell.path_.size()){
				/*  If the path index is smaller than the cell, then we test whether the cell is degenerate in the y-direction. */
				if (degenerate_y){
					/* The cell is y degenerate . We now test which side of the cube the cell is on  */
					if (side_y){
						/* the cell is on the top side */
						/* If the cube path is to the bottom,  then we return false */
						if (cube.path_[i] == 0 || cube.path_[i] == 1)
							return false;
					}
					else{
						/* the cell is on the bottom side */
						/* If the cube path is to the top,  then we return false */
						if (cube.path_[i] == 2 || cube.path_[i] == 3)
							return false;
					}
				}
				/* If the cell is not degenerate in the y direction, we continue on with the for loop */
			}
			/* Otherwise the path index  is the same size or bigger than the cell */
			else if (i< cube.path_.size()){
				/* path index is comparable between the cube and the cel */
				/*  We now test to see if the cube path and the cell path are on opposite y-sides, */
				bool different_sides;
				if (cube.path_[i] == 0 || cube.path_[i] == 1){
					/* The cube is on the BOTTOM side */
					if (cell.path_[i] == 2 || cell.path_[i] == 3){
						/* The cell is on the TOP side */
						different_sides = true;
					}
					else
						different_sides = false;
				}
				else{
					/* The cube is on the TOP side */
					if (cell.path_[i] == 2 || cell.path_[i] == 3){
						/* The cell is on the TOP side */
						different_sides = false;
					}
					else
						different_sides = true;
				}

				/*  If the cube and the cell are on the same side, then we just continue. */
				if (different_sides){
					/* We check to see if the cell is degenerate in the y direction */
					if (degenerate_y){
						outside_y = true;
						/* We check which side of the cube the cell is on */
						if (side_y){
							/* We are on the TOP side */
							/* If the cell.path is on the TOP side, then we return false */
							if (cell.path_[i] == 2 || cell.path_[i] == 3)
								return false;
						}
						else{
							/* We are on the BOTTOM side */
							/* If the cell.path is on the BOTTOM side, then we return false */
							if (cell.path_[i] == 0 || cell.path_[i] == 1)
								return false;
						}
					}
					else
						return false;
				}
			}
		}
		/* We now handle the case where the cube is outside the path of the cell. */
		else{
			/* We test which side of the cube the cell is on  */
			if (side_y){
				/* the cell is on the TOP side */
				if (i < cube.path_.size()){
					/* If the cube path is to the TOP,  then we return false , otherwise continue*/
					if (cube.path_[i] == 2 || cube.path_[i] == 3)
						return false;
				}
				if (i < cell.path_.size()){
					/* If the cell path is to the BOTTOM,  then we return false , otherwise continue*/
					if (cell.path_[i] == 0 || cell.path_[i] == 1)
						return false;
				}
			}
			else{
				/* the cell is on the BOTTOM side  */
				if (i < cube.path_.size()){
					/* If the cube path is to the BOTTOM,  then we return false , otherwise continue*/
					if (cube.path_[i] == 0 || cube.path_[i] == 1)
						return false;
				}
				if (i < cell.path_.size()){
					/* If the cell path is to the TOP,  then we return false , otherwise continue*/
					if (cell.path_[i] == 2 || cell.path_[i] == 3)
						return false;
				}
			}
		}

	}/* end for loop */
	/* If we managed to get through the large 'for' loop without returning false, the nwe get to return true. */
	return true;
}


/*
Obtains a list of SFaces in the complex which have a nonempty intersection with the input SFace.

Needs to have correct CNode*.  splitting path does not matter .

Possible Errors:: A node my have no children and no face info. This will result in a seg fault.
*/
std::vector < SFace > CComplex::Redundant_Cells(SFace in_face){
	/* We find the total cover of the input cell */
	SFace total_cover = Ascend(in_face.pt_node_, in_face.face_number_);
	/* We set the splitting path of in_face to be its LOCAL position, assuming that total_cover is the unit cube */
	in_face.path_ = total_cover.path_;
	/* We clear the splitting path of the total cover */
	total_cover.path_.clear();


	/* We obtain a list of faces which intersect  in_face*/
	std::vector < SFace > face_list = Redundant_Cells_Assisted(in_face, total_cover);
	/* We initialize an output vector */
	std::vector < SFace > output;
	/* We add only the faces from face_list which are actually in the complex */
	for (int i = 0; i < face_list.size(); i++){
		if ((*face_list[i].pt_node_).Check_Cell(face_list[i].face_number_)){
			/* The face is in the complex, */
			output.push_back(face_list[i]);
		}
	}
	return output;

}



/*
***************************
******   PUBLIC     *******
*************************** 
*/





/*
  This function obtains the splitting path of a node 
*/
std::vector< unsigned int > CComplex::get_location( CNode* in_node ){
  /* We initialize a vector which allows us to keep track of how we are ascending the tree */
  std::vector < unsigned int > splitting_path;
  /* If the input node is the root, we return an empty vector */
  if( in_node == &root){
    return splitting_path;
  }
  else{
    /* We initalize a pointer->node called < pt_child >   */
    CNode * pt_child  = in_node;
    /* We continue to ascend the tree until we reach the root. */
    while( !( (pt_child == &root)) ){
      /* We compute the birth order of the child, and add it to the splitting path */
      /*WARNING This way of constructing the splitting path for a node is reversed from the normal method of construction WARNING  */
      splitting_path.push_back((*pt_child).child_number());
      /* We set the child equal to its parent */
      pt_child = (*pt_child).get_parent();
    } 
  }
  /* We initialize an output vector */
  std::vector < unsigned int > output;
  /* Since we wrote down the spiltting path in reverse, we  must reformat our output. */
  while( !(splitting_path.empty())){
    output.push_back(  splitting_path.back());
    splitting_path.pop_back();
  }
  return output;
}


/*
The input is the 'location' of a cube, and we must transverse the tree to find the cooresponding position in the tree

Recall that
* If the input is an empty vector, then the returned cube is the unit cube.
* 0 - left bottom cube
* 1 - right bottom cube
* 2 - left top cube
* 3 - right top cube
*/
CNode* CComplex::get_node(std::vector< unsigned int > location){
	/* We initialize our output to be the address of our root node */
	CNode* pt_output = &root;
	/* We iterate through each position in our location vector */
	for (int i = 0; i < location.size(); i++){
		/* If we encounter a leaf before we get to the end of our location vector, then we need to split the current node */
		if ((*pt_output).leaf_check()){
			(*pt_output).split();
		}
		/* Set the current output pointer ( pt_output ) equal to the appropriate child . */
		pt_output = (*pt_output).get_child(location[i]);
		/* We cycle through, repeating this process until we get to the appropriate depth. */
	}
	return pt_output;
}



/*
This function adds a node to the complex and stores the local approximation.
*/
void CComplex::Add_Cube(std::vector< unsigned int > location, std::vector< bool >  vertices){
	/* We get the faces we need to add to the approximation */
	std::vector< bool > faces = infer_faces(vertices);
	/* We get the pointer to the node which represents the cube we are adding */
	CNode* pt_node = get_node(location);
	/* We store the faces in the node */
	(*pt_node).store_faces(faces);
}


/*
Obtains a list of SFaces in the complex which have a nonempty intersection with the input SFace.

Needs to have full, correct splitting path.  SFace does not need to be part of parent complex. CNode* doesn't matter.

Possible Errors:: A node my have no children and no face info. This will result in a seg fault.
*/
std::vector < SFace > CComplex::Obtain_Carrier( SFace in_face  ){ 
  /* We create an empty vector */
  std::vector < unsigned int > blank;
  /* We use the unit cube as the total cover */ 
  SFace total_cover = { blank , &root, 0 };
  
  return Redundant_Cells_Assisted( in_face , total_cover );
  
}


/* 
Returns the algebraic boundary of the input cell, of the given sign.
false := -1
true  := +1 
*/
std::vector < SFace >   CComplex::Boundary( SFace in_face , bool sign){
  /* We initialize the output */
  std::vector < SFace > output;
  /* We first comput the local boundary of the celll. 
  Afterwards, we need to find the corresponding cells in the complex */
  std::vector  < SFace > local_boundary;
  /* We need a face that we can manipulate */
  SFace example_face = in_face;
  
  /* If the face is a vertex, then we return an empty vector */
  if( in_face.face_number_ >4 )
    return output;
  else if( in_face.face_number_ >0 ){
    /* If the face is a 1-cell, we handle it accordingly */
    if( in_face.face_number_ == 1){
      /* We are dealing with the BOTTOM edge */
      if(sign)
        example_face.face_number_ = 7;
      else
        example_face.face_number_ = 5;
    }
    else if( in_face.face_number_ == 2){
      /* We are dealing with the RIGHT edge */
      if(sign)
        example_face.face_number_ = 8;
      else
        example_face.face_number_ = 7;
    }
    else if( in_face.face_number_ == 3){
      /* We are dealing with the TOP edge */
      if(sign)
        example_face.face_number_ = 8;
      else
        example_face.face_number_ = 6;
    }
    else{
      /* We are dealing with the LEFT edge */
      if(sign)
        example_face.face_number_ = 6;
      else
        example_face.face_number_ = 5;
    }
    local_boundary.push_back( example_face );
  }
  else{
    /* Otherwise the face is a 2-cell */
    if( sign ){
      example_face.face_number_ = 1;
      local_boundary.push_back( example_face );
      example_face.face_number_ = 2;
      local_boundary.push_back( example_face );
    }
    else{
      example_face.face_number_ = 3;
      local_boundary.push_back( example_face );
      example_face.face_number_ = 4;
      local_boundary.push_back( example_face );
    }
  }
  /* Now we must find the actual cell in the complex corresponding to the local boundary */
  /* We are at this point only if the input cell was a 1-cell or a 2-cell */
  for( int i =0 ; i < local_boundary.size(); i++){
    /* If the local boundary cell is actually stored on the input node, then we add it to the output */
    if( (*in_face.pt_node_).Check_Cell(local_boundary[i].face_number_ )){
      output.push_back( local_boundary[i] );
    }
    else{
      /* We get the cells in the complex which intersect the given cell from the local boundary */
      std::vector < SFace > equiv_cells = Redundant_Cells( local_boundary[i] );
      /* We add the appropriate cells to the output */
      for( int j =0 ; j < equiv_cells.size(); j++){
        /* We only want to add cells of the same dimension. 
        This is not a problem if we are computing the boundary of a 1-cell. */
        if( in_face.face_number_ == 0 ){
          /* We are computing the boundary of a 2-cell in this case */
          if( equiv_cells[j].face_number_ >4 ){
            /* the equiv_cell is a vertex. */
            continue;
          }
        }
        output.push_back( equiv_cells[j]);
      }
    }
  }
  return output;
}


void CComplex::Create_CW_Complex( void ){
  /* In order to create a CW Complex, we must remove all large redundant cells */
  /* Initialize the stack of CNode pointers we must prune */
  std::stack< CNode* > Nodes_To_Prune;
  /* We add the pointer to root to the stack */
  Nodes_To_Prune.push(&root);
  while( !(Nodes_To_Prune.empty())){
    /* We remove the top CNode from the stack */
    CNode * current_node = Nodes_To_Prune.top();
    Nodes_To_Prune.pop();
    /* we check if the current node is a leaf */
    if( (*current_node).leaf_check()){
      /* If the current node is a leaf, then we prune the node */
      Prune_Node(current_node);
    }
    else{
      /*If current node is not a leaf, then we add all of its children to the stack */ 
      for( unsigned int i =0; i < 4 ;i++){
        Nodes_To_Prune.push( (*current_node).get_child( i ) );
      }
    }
  }
}


void   CComplex::Print(void ){
  /* In order to Print all the cells in the complex we must iterate through all of the cubes in the complex*/
  /* Initialize the stack of CNode pointers we must add to the cell list */
  std::stack< CNode* > NodeStack;
  /* We add the pointer to root to the stack */
  NodeStack.push(&root);
  while( !(NodeStack.empty())){
    /* We remove the top CNode from the stack */
    CNode * current_node = NodeStack.top();
    NodeStack.pop();
    /* we check if the current node is a leaf */
    if( (*current_node).leaf_check()){
      /* If the current node is a leaf, then we print its information */
      std::cout << " Splitting path :: ";
      /* We first get the splitting path of the cube */
      std::vector < unsigned int > local_path = get_location(current_node);
      /* We print the splitting path of the cube */
      for(int i =0; i< local_path.size(); i++){
        std::cout << local_path[i];
      }
      std::cout << "  ;  ";
      (*current_node).PrintFaces();
    }
    else{
      /*If current node is not a leaf, then we add all of its children to the stack */ 
      for( unsigned int i =0; i < 4 ;i++){
        NodeStack.push( (*current_node).get_child( i ) );
      }
    }
  }
  std::cout << "\n";
  
  
}




