/*
 * CCube.h
 * 
 * This class represents a 2 dimensional dyadic cube by a vector of integers. 
 * This class is used to check whether the input function is well behaved locally, on that dyadic cube.
 * 
 *  Created on: Dec 2, 2013
 *      Author: jono and miro
 */

#ifndef CCUBE_H_
#define CCUBE_H_

#include "SGeometricCube.h"
#include "CPolynomial.h"
#include<vector>
#include <iostream>



class CCube {
private:
	/* location of the cube inside the unit 2D cube
	 * the vector stores consecutive subdivisions
	 * each entry corresponds to the position inside the larger cube.
	 *
	 * If it is empty the cube is the unit cube.
	 * 0 - left bottom cube
	 * 1 - right bottom cube
	 * 2 - left top cube
	 * 3 - right top cube */
        
	std::vector< unsigned int > location_;
        
        bool check_vertices(CPolynomial poly , double threshold ,const  GeometricCube & geom_cube);
        bool strong_derivative_test(std::vector< bool > Jacobian);
        bool weak_derivative_test(CPolynomial poly , std::vector< bool > Jacobian , double threshold , const GeometricCube &geom_cube );
        
        
        /*
         * The next two functions would be better defined as a GeometricCube -member function
        */        
        std::vector< GeometricCube > Get_Weak_Faces( std::vector< bool > Jacobian , const GeometricCube &geom_cube );
        std::vector< GeometricCube > get_vertices(const  GeometricCube & geom_cube);


public:
  
  
	CCube();
	/* Copy constructor */
	CCube( const CCube& cube );
	/* Location constructor copies the location to the new cube */
	CCube( std::vector< unsigned int > location );
	virtual ~CCube();

	/* Returns a vector with four cubes which form the subdivision of the original cube */
	std::vector< CCube > Subdivide();
	/* returns the geometric position of the cube */
        GeometricCube  GetGeometricRepresentation();
	/* prints the location vector */
	void Print();
        int verify( CPolynomial poly , double threshold );
        
        /* These two functions are used to add a cube to the complex */
        std::vector< unsigned int > get_location( void ){ return location_ ;};
        /* Also note that local_approximation is similar to check_vertices . */ 
        std::vector< bool > local_approximation( CPolynomial poly , double threshold );
        
};

#endif /* CCUBE_H_ */
