/*
 * CFiltration.h
 *
 * This class is derived from the base class Complex. 
 * It represents a CW complex on the entire unit square and is used to export a filtered algebraic chain complex.
 *  Created on: Dec 6, 2013
 *      Author: Jono
 */

#ifndef CFILTRATION_H_
#define CFILTRATION_H_

#include "CComplex.h"
#include <fstream>
#include <string>
#include <sstream>
#include <iostream>

class CFiltration : public CComplex {
  private:
    /* In addition to the tree structure, we store a list of the cells, first sorted by dimension. */
    std::vector < SFace >  Zero_Cell_List;
    std::vector < SFace >  One_Cell_List;
    std::vector < SFace >  Two_Cell_List;
    
    
    bool            Equal_Support( SFace  cell, SFace mu) ; 
    bool            Containment( SFace  mu , SFace  lambda  );
    
    void            Update_Cell( SFace lambda);                      
    void            Update_Minimal_Birth( SFace lambda , std::vector< SFace > cover );
	void            Update_Complex(void);
    
    void            Add_Cell_To_Complex( SFace cell, int threshold  );
    void            Add_Cube_To_List( CNode* node  );
    std::string     Write_Cell( SFace cell );
    void            Build_List( void );
    SFace           Refine(SFace  lambda, SFace carrier);
        
    bool            Unique_Carrier( SFace cell, std::vector < SFace >  cover ) ;


  public:
    void            Initialize( unsigned int max_steps);
    void            Add_Complex( CComplex & complex , int threshold);
    void            Export( void );
     
};

#endif /* CFILTRATION_H_ */

