//============================================================================
// Name        : Tutorial.cpp
// Author      : Jono Jaquette
// Version     :  1.2   July 31 2014 - August 5 2014
// Copyright   : Your copyright notice
// Description : A tutorial for how to compute persistent homology of 2-dimensional functions.
//============================================================================


/*
CApproximation.cpp includes all of the other classes, structures and libraries needed by this program. 
One such file we need to include is:
<boost/numeric/interval.hpp>
In order to compile Tutorial.cpp, you will need to include a path to your boost library.
*/


#include "CBasicPolynomial.cpp"
#include "CPolynomial.cpp"
#include "CCube.cpp"
#include "CApproximation.cpp"
#include "CNode.cpp"
#include "CComplex.cpp"
#include "CFiltration.cpp"



int main() {
	
	/* Initalizes the object which will store a filtered CW complex */
	CFiltration filter;
	/* Initializes the object which will produce the filtered CW complex*/
    CApproximation approx;
  


    /* Defines the smallest thresholds in the filtration */
	double bottom_threshold = -.4;
	/* Defines the spacing in between thresholds in the filtration */
	double step_size        =   .1;
	/* Defines the total number thresholds included in the filtration */
	unsigned int max_steps  =   7;
	/* Decides if we are studying the sub-level set (true) or the super-level set (false). */
	bool sublevel = true;





	/* Calls the function which produces the filtered CW complex (induced by function super-level sets) */
	approx.Verified_Filtration(bottom_threshold, step_size, max_steps, sublevel, filter);



	std::cout << "\nProgram is terminating ... \n";

  
 
  return 0;
}
